/*
*
* Template Numerical Toolkit (TNT)
*
* Mathematical and Computational Sciences Division
* National Institute of Technology,
* Gaithersburg, MD USA
*
*
* This software was developed at the National Institute of Standards and
* Technology (NIST) by employees of the Federal Government in the course
* of their official duties. Pursuant to title 17 Section 105 of the
* United States Code, this software is not subject to copyright protection
* and is in the public domain. NIST assumes no responsibility whatsoever for
* its use by other parties, and makes no guarantees, expressed or implied,
* about its quality, reliability, or any other characteristic.
*
*/


#ifndef TNT_ARRAY2D_UTILS_H
#define TNT_ARRAY2D_UTILS_H

#include <cstdlib>
#include <cassert>

namespace TNT
{


template <class T>
std::ostream& operator<<(std::ostream &s, const Array2D<T> &A)
{
    int M=A.dim1();
    int N=A.dim2();

    s << M << " " << N << "\n";

    for (int i=0; i<M; i++)
    {
        for (int j=0; j<N; j++)
        {
            s << A[i][j] << " ";
        }
        s << "\n";
    }


    return s;
}

template <class T>
std::istream& operator>>(std::istream &s, Array2D<T> &A)
{

    int M, N;

    s >> M >> N;

	Array2D<T> B(M,N);

    for (int i=0; i<M; i++)
        for (int j=0; j<N; j++)
        {
            s >>  B[i][j];
        }

	A = B;
    return s;
}


//------------------------- modified - if not same size, return intersection
template <class T>
Array2D<T> operator+(const Array2D<T> &A, const Array2D<T> &B){
	int m = min(A.dim1(), B.dim1());
	int n = min(A.dim2(), B.dim2());
	Array2D<T> C(m,n);
	for( int i=0; i<m; i++ )
		for( int j=0; j<n; j++ )
			C[i][j] = A[i][j] + B[i][j];
	return C;
	}

template <class T>
Array2D<T> operator-(const Array2D<T> &A, const Array2D<T> &B){
	int m = min(A.dim1(), B.dim1());
	int n = min(A.dim2(), B.dim2());
	Array2D<T> C(m,n);
	for( int i=0; i<m; i++ )
		for( int j=0; j<n; j++ )
			C[i][j] = A[i][j] - B[i][j];
	return C;
	}

template <class T>
Array2D<T> operator*(const Array2D<T> &A, const Array2D<T> &B){
	int m = min(A.dim1(), B.dim1());
	int n = min(A.dim2(), B.dim2());
	Array2D<T> C(m,n);
	for( int i=0; i<m; i++ )
		for( int j=0; j<n; j++ )
			C[i][j] = A[i][j] * B[i][j];
	return C;
	}

template <class T>
Array2D<T> operator/(const Array2D<T> &A, const Array2D<T> &B){
	int m = min(A.dim1(), B.dim1());
	int n = min(A.dim2(), B.dim2());
	Array2D<T> C(m,n);
	for( int i=0; i<m; i++ )
		for( int j=0; j<n; j++ )
			C[i][j] = B[i][j]==0 ? 0 : (A[i][j]/B[i][j]);
	return C;
	}

//------------------------- modified - if not same size, calculate as much as possible
template <class T>
Array2D<T> & operator+=( Array2D<T> &A, const Array2D<T> &B ){
	int m = min( A.dim1(), B.dim1() );
	int n = min( A.dim2(), B.dim2() );
	for( int i=0; i<m; i++ )
		for( int j=0; j<n; j++ )
			A[i][j] += B[i][j];
	return A;
	}

template <class T>
Array2D<T> & operator-=( Array2D<T> &A, const Array2D<T> &B ){
	int m = min( A.dim1(), B.dim1() );
	int n = min( A.dim2(), B.dim2() );
	for( int i=0; i<m; i++ )
		for( int j=0; j<n; j++ )
			A[i][j] -= B[i][j];
	return A;
	}

template <class T>
Array2D<T> & operator*=( Array2D<T> &A, const Array2D<T> &B ){
	int m = min( A.dim1(), B.dim1() );
	int n = min( A.dim2(), B.dim2() );
	for( int i=0; i<m; i++ )
		for( int j=0; j<n; j++ )
			A[i][j] *= B[i][j];
	return A;
	}

template <class T>
Array2D<T> & operator/=( Array2D<T> &A, const Array2D<T> &B ){
	int m = min( A.dim1(), B.dim1() );
	int n = min( A.dim2(), B.dim2() );
	for( int i=0; i<m; i++ )
		for( int j=0; j<n; j++ )
			if( B[i][j]!=0 ) 
				A[i][j] /= B[i][j];
	return A;
	}

/**
    Matrix Multiply:  compute C = A*B, where C[i][j]
    is the dot-product of row i of A and column j of B.


    @param A an (m x n) array
    @param B an (n x k) array
    @return the (m x k) array A*B, or a null array (0x0)
        if the matrices are non-conformant (i.e. the number
        of columns of A are different than the number of rows of B.)

extension by JTB - Versions of matmult : 
( 2d * 2d = 2d ) 
void       matmult( int iI, intiJ, int iK,  T ** A,  T ** B, T**C, bool bSum=false )
Array2D<T> matmult( int iI, intiJ, int iK,  T ** A,  T ** B )
Array2D<T> matmult(  Array2D<T> &A,  Array2D<T> &B )

( 2d * 1d = 1d ) 
void       matmult( int iI, int iJ,  T ** A,  T * B, T * C, bool bSum=false )
Array1D<T> matmult( int iI, int iJ,  T ** A,  T * B )
Array1D<T> matmult(  Array2D<T> &A,  Array1D<T> &B )

( 1d * 2d = 1d ) 
void       matmult( int iI, int iJ,  T * A,  T ** B, T * C, bool bSum=false )
Array1D<T> matmult( int iI, int iJ,  T * A,  T ** B ) 
Array1D<T> matmult(  Array1D<T> &A,  Array2D<T> &B )

*/

// c[][] = a[][] * b[][]  -or- c[][] += a[][] * b[][]
template <class T>
void matmult( int iI, int iJ, int iK,  T** A,  T** B, T** C, bool bSum=false ) { 
    for( int i=0; i<iI; i++ )
        for( int k=0; k<iK; k++ ){
            T sum = 0;
            for( int j=0; j<iJ; j++ )
                sum += A[i][j] * B [j][k];
            if( bSum ) 
				C[i][k] += sum;
			else 
				C[i][k] = sum;
			}
	}

// return A2d = a[][] * b[][]  
template <class T>
Array2D<T> matmult( int iI, int iJ, int iK,  T ** A,  T ** B ) { 
    Array2D<T> C(iI,iK);
    matmult( iI, iJ, iK, ( T **)A, ( T **)B, (T **)C );
    return C;
    }

// return A2d = A2d * A2d
template <class T>
Array2D<T> matmult(  Array2D<T> &A,  Array2D<T> &B ){
	return matmult( A.dim1(), min(A.dim2(),B.dim1()), B.dim2(), A, B );
	}

// c[][] = a[][] * b[]  -or- c[][] += a[][] * b[]
template <class T>
void matmult( int iI, int iJ,  T ** A,  T * B, T * C, bool bSum=false ){
    for( int i=0; i<iI; i++ ) { 
		T dSum = 0.0;
        for( int j=0; j<iJ; j++)
			dSum += A[i][j] * B [j];
		if( bSum ) 
			C[i] += dSum;
		else
			C[i] = dSum;
		}
	}

// return A1d = a[][] * b[]
template <class T>
Array1D<T> matmult( int iI, int iJ,  T ** A,  T * B ) { 
    Array1D<T> C(iI);
    matmult( iI, iJ, (double **)A, (double *)B, (double *)C );
    return C;
    }

// return A1d = A2d * A1d
template <class T>
Array1D<T> matmult(  Array2D<T> &A,  Array1D<T> &B ) { 
	return matmult( A.dim1(), min( A.dim2(), B.dim()), (T **)A, (T *)B );
	}

// c[][] = a[] * b[][]  -or- c[][] += a[] * b[][]
template <class T>
void matmult( int iI, int iJ,  T * A,  T ** B, T * C, bool bSum=false ){
    for( int j=0; j<iJ; j++) { 
		T dSum = 0.0;
	    for( int i=0; i<iI; i++ )
			dSum += A[i] * B [i][j];
		if( bSum ) 
			C[j] += dSum;
		else
			C[j] = dSum;
		}
	}

// return A1d = a[] * b[][]
template <class T>
Array1D<T> matmult( int iI, int iJ,  T * A,  T ** B ) { 
    Array1D<T> C(iJ);
    matmult( iI, iJ, (double *)A, (double **)B, (double *)C );
    return C;
    }

// return A1d = A2d * A1d
template <class T>
Array1D<T> matmult(  Array1D<T> &A,  Array2D<T> &B ) { 
	return matmult( min( A.dim(), B.dim1()), B.dim2(), (T *)A, (T **)B );
	}

//----- swap rows i0 and i1 
template <class T>
void SwapRows( int i0, int i1, int iCols, T ** adA ){ 		
	for( int iCol=0; iCol<iCols; ++iCol ) 
		Swap( adA[i0]+iCol, adA[i1]+iCol );
	}

template <class T>
void Array2D<T>::SwapRows( int i0, int i1 ){
	#ifdef TNT_BOUNDS_CHECK
	if( i0<0 || i0>=m_ || i1<0 || i1>=m_ || i0==i1 ) 
		return;
	#endif
	
	T d;
	for( int iCol=0; iCol<n_; ++iCol ) {
		d = v_[i0][iCol];
		v_[i0][iCol] = v_[i1][iCol];
		v_[i1][iCol] = d;
		}
	}
	
//----- swap columns i0 and i1 
template <class T>
void Array2D<T>::SwapCols( int i0, int i1 ){
	#ifdef TNT_BOUNDS_CHECK
	if( i0<0 || i0>=n_ || i1<0 || i1>=n_ || i0==i1 ) 
		return;
	#endif
	
	T d;
	for( int iRow=0; iRow<n_; ++iRow ) {
		d = v_[iRow][i0];
		v_[iRow][i0] = v_[iRow][i1];
		v_[iRow][i1] = d;
		}
	}

//-- extension by JTB - matrix inversion by Gauss Jordan elimination 
template <class T>
bool MatrixInvert( int iDim, T ** pdA ) { 
	const double MACHEP = 2.220446049250313E-16;     //{ 2^(-52) } (type must be double or better)
	int i,j,k;
	T dPiv, d;
	int * aPivRow = new int[ iDim ];
	int * aPivCol = new int[ iDim ];
	for( k=0; k<iDim; ++k ) { 
		//-- Search for largest pivot in A[k..n][k..n]
		dPiv = pdA[aPivRow[k]=k][aPivCol[k]=k];
		for( i=k; i<iDim; ++i ) 
			for( j=k; j<iDim; ++j ) 
				if( fabs( pdA[i][j] ) > fabs(dPiv) ) 
					dPiv = pdA[aPivRow[k]=i][aPivCol[k]=j];
					
		//-- Pivot too weak ==> quasi-singular matrix
		if( fabs( dPiv ) < MACHEP ) {
			delete [] aPivRow;
			delete [] aPivCol;
			return false;
			}
		
		//-- Exchange current row (K) with pivot row
		SwapRows( k, aPivRow[k], iDim, (double **)pdA );

		//-- Exchange current column (K) with pivot column
		SwapCols( iDim, k, aPivCol[k], (double **)pdA );

		//-- Transform pivot row 
		pdA[k][k]=1.0;
		for( j=0; j<iDim; ++j ) 
			pdA[k][j]/=dPiv;

		//-- Transform other rows
		for( i=0; i<iDim; ++i ) 
			if( i != k ) { 
				d=pdA[i][k];
				pdA[i][k]=0.0;
				for( j=0; j<iDim; ++j ) 
					pdA[i][j] -= d * pdA[k][j];
				}
		}

	//-- Rearrange inverse matrix 
	for( i=iDim-1; i>=0; --i ) 
		SwapRows( i, aPivCol[i], iDim, (double **)pdA );
	for( j=iDim-1; j>=0; --j ) 
		SwapCols( iDim, j, aPivRow[j], (double **)pdA );

	delete [] aPivRow;
	delete [] aPivCol;
	return true;
	}	

template <class T>
bool MatrixInvert( Array2D<T> & A ) { 
	return MatrixInvert( A.dim1(), (T **) A );
	}
	

//-- sum product of 2 arrays (not dot product) : d = sum( a[i][j] * b[i][j] ) 
template <class T>
T SumProduct( int iI, int iJ, T ** adA, T ** adB ){
	T dSum = 0.0;
	for( int i=0; i<iI; i++)
		for( int j=0; j<iJ; j++ )
			dSum += adA[i][j] * adB[i][j];
	return dSum;
	}


//------------------------------------
// Getting into some pretty esoteric territory here, but it fits better in TNT than elsewhere...
//-- extension by JTB - multiply matrices which do not fill allocation A[I,J] * B[J,K] 
// The [i,k] are inserted into positions of C indicated by the arrays of rows and cols used.
// for example, with a 3x3 multiply, aiRows={1,7,9} and aiCols={2,3,4} then the result array
// is placed in C[1,2] C[1,3] C[1,4] C[7,2] C[7,3] C[7,4] C[9,2] C[9,3] C[9,4].
// This is specific to the Q matrix needs of Mil.
template <class T>
void matmultintomap( int iI, int iJ, int iK, int * aiRows, int * aiCols, 
					  T ** A, T ** B, T ** C, bool bSum=false ){
    for( int i=0; i<iI; ++i ) 
		for( int k=0; k<iK; ++k ) { 
			T tSum = 0.0;
			for( int j=0; j<iJ; ++j ) 
				tSum += A[i][j]*B[j][k];
			if( bSum ) 
				C[aiRows[i]][aiCols[k]]+=tSum;
			else
				C[aiRows[i]][aiCols[k]]=tSum;
			}
	}

template <class T>
void vecmultintomap( int iRows, int iCols, int * aiRows, int * aiCols, 
					  T * A, T * B, T ** C, bool bSum=false ){
	if( bSum ) 
	    for( int i=0; i<iRows; ++i ) 
			for( int k=0; k<iCols; ++k )  
				C[aiRows[i]][aiCols[k]]+=A[i]*B[k];
	else
	    for( int i=0; i<iRows; ++i ) 
			for( int k=0; k<iCols; ++k )  
				C[aiRows[i]][aiCols[k]]=A[i]*B[k];
	}

//-- extract as a new array the rows and columns of A specified in aiRows and aiCols
/*
template <class T>
Array2D<T> matextract( int iI, int iJ, int * aiRows, int * aiCols, const Array2D<T> &A ) { 
	Array2D<T> C( iI, iJ );
	for( int i=0; i<iI; ++i )
		for( int j=0; j<iJ; ++j ) 
			C[i][j]=A[aiRows[i]][aiCols[j]];
	return C;
	}
*/	

template <class T>
void matextract( int iI, int iJ, const int * aiRows, const int * aiCols, T ** adA, T ** adResult ) { 
	for( int i=0; i<iI; ++i )
		for( int j=0; j<iJ; ++j ) 
			adResult[i][j]=adA[aiRows[i]][aiCols[j]];
	}

template <class T>
Array2D<T> matextract( int iI, int iJ, const int * aiRows, const int * aiCols, T ** adA ) { 
	Array2D<T> C( iI, iJ );
	// matextract( iI, iJ, aiRows, aiCols, (double **) C );
	for( int i=0; i<iI; ++i )
		for( int j=0; j<iJ; ++j ) 
			C[i][j]=adA[aiRows[i]][aiCols[j]];
	return C;
	}

//-- extract a vector of the positions of elements in a[] which match/differ with tFind
template <class T>
Array1D<int> matextractmap( int iLen, T * a, T tFind, bool bMatch ) { 
	int iCount=0;
	for( int i=0; i<iLen; ++i ) 
		if( ( a[i]==tFind ) == bMatch ) 
			++iCount;
	Array1D<int> C( iCount );
	iCount=0; 
	for( int i=0; i<iLen; ++i ) 
		if( ( a[i]==tFind ) == bMatch ) 
			C[iCount++]=i;
	return C;
	}

//-- same as above with 2 tests : bMatch=true:(=1 or =2)  bMatch=false:(!=1 and !=2) 
/*
template <class T>
Array1D<int> mapextract( int iLen, T * a, T tFind1, T tFind2, bool bMatch ) { 
	int iCount=0;
	for( int i=0; i<iLen; ++i ) 
		if( bMatch ? ( a[i]==tFind1 || a[i]==tFind2) : ( a[i]!=tFind1&&a[i]!=tFind2 )  ) 
			++iCount;
	Array1D<int> aMap( iCount );
	iCount=0; 
	for( int i=0; i<iLen; ++i ) 
		if( bMatch ? ( a[i]==tFind1 || a[i]==tFind2) : ( a[i]!=tFind1&&a[i]!=tFind2 )  ) 
			a[iCount++]=i;
	}
*/

template <class T>
void matcopy( int iI, int iJ, T ** adSource, T ** adDestination ) { 
	for( int i=0; i<iI; ++i )
		for( int j=0; j<iJ; ++j ) 
			adDestination[i][j]=adSource[i][j];
	}

//-- Multiply a vector and an extract of an array to produce a vector
// a[] * b[rows[]][cols[]] = C[]
// a[rows[]][cols[]] * b[] = C[]
template <class T>
void matextractmult( int iI, int iJ, T * adA, const int * aiRows, const int * aiCols, T ** adB, T * adC ) { 
	for( int j=0; j<iJ; ++j ) {
		T Sum = 0.0;
		int iCol = aiCols[j];
		for( int i=0; i<iI; ++i )
			Sum += adA[i] * adB[aiRows[i]][iCol];
		adC[j] = Sum;
		}
	}
	
template <class T>
void matextractmult( int iI, int iJ, const int * aiRows, const int * aiCols, T ** adA, T * adB, T * adC ) { 
	for( int i=0; i<iI; ++i ) { 
		int iRow = aiRows[i];
		T Sum = 0.0;
		for( int j=0; j<iJ; ++j )
			Sum += adA[iRow][aiCols[j]] * adB[j];
		adC[i] = Sum;
		}
	}

template <class T>
bool Spectrum( Array2D<T> & aA, T * adEigVal, T ** adEigVec, T ** adEigVecInv ){
	int iDim1 = aA.dim1();
	int iDim2 = aA.dim2();

	//-- eigenvalues calculation 
	JAMA::Eigenvalue<T> oQEigVal( aA );

	//-- copy eigenvalues to output and check them
	for( int i=0; i<iDim1; ++i ) { 
		// copy real part to array real part 
		if( (adEigVal[i]=oQEigVal.d[i]) > 1.0e-5 ) {
			// TODO - report positive eigenvalues found
			return false;					
			}
			
		if( fabs(oQEigVal.e[i]) > 1.0e-5 ) {
			// TODO - report complex eigenvalues 
			return false;		
			}
		}

	//-- eigenvectors
	matcopy( iDim1, iDim2, (T **) oQEigVal.V, (T **) adEigVec );
	
	//-- eigenvectors inverse
	matcopy( iDim1, iDim2, (T **) adEigVec, (T **) adEigVecInv );
	return MatrixInvert( iDim1, adEigVecInv );
	}

//-- R[m][m] = A[m][m] * B[m][m] * C[m][m]
template <class T>
void MxMxM( int m, T**A, T**B, T**C, Array2D<T> & R ) { 
	R=0.0;
	for( int i=0; i<m; ++i ) {
		for( int j=0; j<m; ++j ) {
			double d=0.0;
			for( int k=0; k<m; ++k ) 
				d += A[i][k] * B[k][j];
			for( k=0; k<m; ++k ) 
				R[i][k] += d * C[j][k];
			}
		}
	}

//-- R[mapped aiRows[m]][mapped aiCols[m]] = A[m][m] * B[m][m] * C[m][m]
template <class T>
void MxMxMIntoMap( int m, T**A, T**B, T**C, int * aiRows, int * aiCols, T**R ) { 
	for( int i=0; i<m; ++i ) {
		for( int j=0; j<m; ++j ) {
			double d=0.0;
			for( int k=0; k<m; ++k ) 
				d += A[i][k] * B[k][j];
			for( k=0; k<m; ++k ) 
				R[aiRows[i]][aiCols[k]] += d * C[j][k];
			}
		}
	}
	
#ifdef _DEBUG
template <class T>
void OutputDebugArray( const Array2D<T> & a, const char * pchName ) { 
	char buffer[256];
	for( int i=0; i<a.dim1() && i<9; ++i ) {
		sprintf( buffer, "%s{", i==0 ? pchName : "\t\t" );
		for( int j=0; j<a.dim2() && j<12; ++j ) 
			sprintf( buffer+strlen(buffer), "\t%.9f", double(a[i][j]) );
		if( a.dim2()>=12 ) 
			strcat( buffer, "..." );
		strcat( buffer, "}\n" );
		OutputDebugString( buffer );
		}
	if( a.dim1()>=9 ) { 
		sprintf(buffer, "\t\t... ... (%d Rows)\n", a.dim1() );
		OutputDebugString( buffer );
		}
	}
#endif


} // namespace TNT


#endif
